/*******************************************************************************
' Author: Dave Hein
' Version 0.54
' Copyright (c) 2012
' See end of file for terms of use.
'******************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "spinsim.h"

#define STATE_IDLE   0
#define STATE_DEVICE 1
#define STATE_ADDR1  2
#define STATE_ADDR2  3
#define STATE_WRITE  4
#define STATE_READ   5

#define CLOCK_HELD_LOW  (!scl_prev && !scl)
#define CLOCK_HELD_HIGH ( scl_prev &&  scl)
#define CLOCK_FALLING   ( scl_prev && !scl)
#define DATA_RISING     (!sda_prev &&  sda)
#define DATA_FALLING    ( sda_prev && !sda)

#define DEVICE_ADDR 0xA0
#define ACKNOWLEDGE 0

extern int32_t pin_val_a;
extern int32_t eeprom;

static int32_t scl_prev = 1;
static int32_t sda_prev = 1;
static int32_t state = STATE_IDLE;
static int32_t bitcount = 0;
static int32_t value = 0;
static int32_t address = 0;
static int32_t drivepin = 0;
static int32_t driveval = 0;

static unsigned char memory[256*256];

void CheckEEProm()
{
    int32_t scl = (pin_val_a >> 28) & 1;
    int32_t sda = (pin_val_a >> 29) & 1;

    if (!eeprom) return;

    // START
    if (!drivepin && CLOCK_HELD_HIGH && DATA_FALLING)
    {
        bitcount = 0;
        state = STATE_DEVICE;
        value = 0;
        drivepin = 0;
    }

    // STOP
    else if (!drivepin && CLOCK_HELD_HIGH && DATA_RISING)
    {
        state = STATE_IDLE;
        drivepin = 0;
    }

    // CLOCK DEVICE
    else if (state && CLOCK_FALLING)
    {
        drivepin = 0;

        // DEVICE ADDRESS
        if (state == STATE_DEVICE)
        {
            if (bitcount) value = (value << 1) | sda_prev;
            bitcount++;
            if (bitcount == 9)
            {
                if ((value & 0xf0) == DEVICE_ADDR)
                {
                    drivepin = 1;
                    driveval = ACKNOWLEDGE;
                    bitcount = 0;
                    if (value & 1)
                        state = STATE_READ;
                    else
                        state = STATE_ADDR1;
                }
                else
                {
                    state = STATE_IDLE;
                }
            }
        }

        // ADDRESS
        else if (state == STATE_ADDR1 || state == STATE_ADDR2)
        {
            if (bitcount) address = (address << 1) | sda_prev;
            else if (state == STATE_ADDR1) address = 0;
            bitcount++;
            if (bitcount == 9)
            {
                state++;
                drivepin = 1;
                driveval = ACKNOWLEDGE;
                bitcount = 0;
            }
        }

        // WRITE DATA
        else if (state == STATE_WRITE)
        {
            if (bitcount) value = (value << 1) | sda_prev;
            bitcount++;
            if (bitcount == 9)
            {
                memory[address] = value;
                address = (address + 1) & 0xffff;
                drivepin = 1;
                driveval = ACKNOWLEDGE;
                bitcount = 0;
                eeprom = 2;
            }
        }

        // READ DATA
        else if (state == STATE_READ)
        {
            if (bitcount == 0)
            {
                if (sda_prev)
                    state = STATE_IDLE;
                else
                {
                    value = memory[address];
                    drivepin = 1;
                    driveval = (value >> 7) & 1;
                    value <<= 1;
                    bitcount++;
                    address = (address + 1) & 0xffff;
                }
            }
            else if (bitcount < 8)
            {
                drivepin = 1;
                driveval = (value >> 7) & 1;
                value <<= 1;
                bitcount++;
            }
            else
            {
                bitcount = 0;
                drivepin = 0;
            }
        }
    }
    if (drivepin)
    {
        pin_val_a = (pin_val_a & (~(1 << 29))) | (driveval << 29);
        sda = driveval;
    }
    scl_prev = scl;
    sda_prev = sda;
}

static FILE *OpenFile(char *fname, char *mode)
{
    FILE *file = fopen(fname, mode);
    if (!file)
    {
        printf("Could not open %s\n", fname);
        spinsim_exit(1);
    }
    return file;
}

void EEPromInit(char *fname)
{
    FILE *file;

    if (!eeprom) return;

    //printf("EEPromInit(%s)\n", fname);

    if (fname)
    {
        file = fopen("eeprom.dat", "r");
        if (file)
        {
            fread(memory, 1, 65536, file);
            fclose(file);
            memset(memory, 0, 32768);
        }
        else
            memset(memory, 0, 65536);
        file = OpenFile(fname, "r");
        fread(memory, 1, 65536, file);
        fclose(file);
        file = OpenFile("eeprom.dat", "w");
        fwrite(memory, 1, 65536, file);
        fclose(file);
        return;
    }
    file = OpenFile("eeprom.dat", "r");
    fread(memory, 1, 65536, file);
    fclose(file);
}

void EEPromClose(void)
{
    FILE *file;

    if (eeprom == 2)
    {
        file = OpenFile("eeprom.dat", "w");
        fwrite(memory, 1, 65536, file);
        fclose(file);
    }
}

void EEPromCopy(char *mem)
{
    if (!eeprom) return;

    //printf("EEPromCopy\n");
    memcpy(mem, memory, 32768);
}
/*
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
*/
